import pandas as pd
from sqlalchemy import (Boolean, Column, DateTime, Float, Integer, MetaData,
                        String, Table, create_engine)
from sqlalchemy.ext.automap import automap_base
# from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

import observations as obs

engine = create_engine(
    'postgresql://postgres:Io0bgSd9@127.0.0.1:5432/test_men2', echo=True, paramstyle="format")


ts_table_name = '"StandpipeMeas"'
pipe_id = 581
sql = """ SELECT * FROM {} WHERE "PipeId" = {};""".format(
    ts_table_name, pipe_id)
#sql = """ SELECT * FROM {};""".format('"ObservationWells"')


raw_series = pd.read_sql(
    sql, con=engine, index_col='DateTime', parse_dates=[0])
# %%
Base = automap_base()
Base.prepare(engine, reflect=True)

DBSession = sessionmaker(bind=engine)
session = DBSession()


obs_well_id = session.query(
    Base.classes.ObservationWells.ObsWellId).all()[0][0]

obs_dic = session.query(Base.classes.ObservationWells).get(
    obs_well_id).__dict__
loc_dic = session.query(Base.classes.Locations).get(
    obs_dic['LocationId']).__dict__
sp_dic = session.query(Base.classes.Standpipes).filter(
    Base.classes.Standpipes.ObsWellId == obs_well_id)

obs_dic.pop('_sa_instance_state')

o = obs.GroundwaterObs(x=loc_dic.pop('XCoordinate'), y=loc_dic.pop('YCoordinate'),
                       maaiveld=loc_dic.pop('SurfaceLevel'), )



import numpy as np
import pandas as pd
import scipy.interpolate as intp

from . import accessor


def get_model_layer_z(z, zvec, left=-999, right=999,
                      verbose=False):
    """get index of model layer based on elevation. Assumptions:

        - the highest value in zvec is the top of model layer 0.
        - if z is equal to the bottom of a layer, the model layer above that
        layer is assigned.

    Parameters
    ----------
    z : int, float
        elevation.
    zvec : list, np.array
        elevations of model layers. shape is nlay + 1
    left : int, optional
        if z is below the lowest value in zvec, this value is returned. 
        The default is -999.
    right : TYPE, optional
        if z is above the highest value in zvec, this value is returned. 
        The default is 999.

    Returns
    -------
    int
        model layer.
        
    Examples
    --------
    >>> zvec = [0, -10, -20, -30]
    >>> get_model_layer_z(-5, zvec)
    0
    
    >>> get_model_layer_z(-25, zvec)
    2
    
    >>> get_model_layer_z(-50, zvec)
    -999
    
    >>> get_model_layer_z(100, zvec)
    999
    
    >>> get_model_layer_z(-20, zvec)
    1
    """
    # make sure zvec is descending
    zvec = np.sort(zvec)[::-1]
    if z in zvec:
        z+= 1e-10
    lay = int(np.interp(z, zvec[::-1], np.arange(len(zvec))[::-1],
              left=left, right=right))
    
    return lay

def check_if_var_is_invalid(var):
    
    if var is None:
        return True
    elif np.isnan(var):
        return True
    
    return False

def get_modellayer_from_filter(ftop, fbot, zvec, left=-999, right=999,
                               verbose=False):
    """
    

    Parameters
    ----------
    ftop : int or float
        top of filter.
    fbot : int or float
        bottom of filter, has to be lower than ftop.
    zvec : list or np.array
        elevations of the modellayers at the location of the filter.
    left : int, optional
        value to return if filter is below the modellayers. 
        The default is -999.
    right : int, optional
        value to return if filter is above the modellayers. 
        The default is-999.
    verbose : bool, optional
        DESCRIPTION. The default is False.

    Raises
    ------
    ValueError
        raised if something unexpected happens.

    Returns
    -------
    int or np.nan
        modellayer.
        
    examples
    --------
    >>> zvec = [0, -10, -20, -30, -40]
    >>> get_modellayer_from_filter(-5, -7, zvec)
    0
    
    >>> get_modellayer_from_filter(-25, -27, zvec)
    2
    
    >>> get_modellayer_from_filter(-15, -27, zvec, verbose=True)
    2
    
    >>> get_modellayer_from_filter(-5, -27, zvec, verbose=True)
    1
    
    >>> get_modellayer_from_filter(-5, -37, zvec, verbose=True)
    nan
    
    >>> get_modellayer_from_filter(15, -5, zvec, verbose=True)
    0
    
    >>> get_modellayer_from_filter(15, 5, zvec, verbose=True)
    999
    
    >>> get_modellayer_from_filter(-55, -65, zvec, verbose=True)
    -999
    
    >>> get_modellayer_from_filter(15, -65, zvec, verbose=True)
    nan
    
    >>> get_modellayer_from_filter(None, -7, zvec, verbose=True)
    0
    
    >>> get_modellayer_from_filter(None, None, zvec, verbose=True)
    nan
    
    """

    zvec = np.sort(zvec)[::-1]
    ftop_invalid = check_if_var_is_invalid(ftop)
    fbot_invalid = check_if_var_is_invalid(fbot)
    if ftop_invalid and fbot_invalid:
        return np.nan
    elif fbot_invalid:
        lay_ftop = get_model_layer_z(ftop, zvec,
                                     left=left, right=right)
        return lay_ftop
    elif ftop_invalid:
        lay_fbot = get_model_layer_z(fbot, zvec,
                                     left=left, right=right)
        return lay_fbot
    
    if ftop<fbot:
        if verbose:
            print('filter top below filter bot, switch top and bot')
        fbot, ftop = ftop, fbot
        
    lay_ftop = get_model_layer_z(ftop, zvec,
                                 left=left, right=right)
    lay_fbot = get_model_layer_z(fbot, zvec,
                                 left=left, right=right)
    
    # Piezometer in layer in which majority of screen is located
    if lay_fbot == lay_ftop:
        return lay_fbot
    
    else:
        if lay_fbot==left and lay_ftop==right:
            if verbose:
                print('filter spans all layers. '
                      'return nan')
            return np.nan
        elif lay_ftop==right:
            if verbose:
                print('filter top higher than top layer. '
                      'selected layer {}'.format(lay_fbot))
            return lay_fbot
        
        elif lay_fbot==left:
            if verbose:
                print('filter bot lower than bottom layer. '
                      'selected layer {}'.format(lay_ftop))
            return lay_ftop
        
        if verbose:
            print("filter crosses layer boundary:")
            print("-layers: {0}, {1}".format(lay_ftop, lay_fbot))
            print("-layer elev in between: {0:.2f} - {1:.2f}".format(zvec[lay_fbot+1], zvec[lay_ftop]))
            print("-filter elev in between: {0:.2f} - {1:.2f}".format(fbot, ftop)) 
        
        if lay_fbot - lay_ftop > 2:
            if verbose:
                print(f"Piezometer filter spans {lay_fbot - lay_ftop +1} layers."
                      " return nan")
            return np.nan
        
        elif lay_fbot - lay_ftop == 2:  # if filter spans 3 layers
            if verbose:
                print(f"Piezometer filter spans {lay_fbot - lay_ftop +1} layers."
                      f" Use middle layer {lay_ftop+1}")
            return lay_ftop + 1  # use middle layer
        
        # check which layer has the biggest length of the filter 
        length_lay_bot = zvec[lay_fbot] - fbot
        length_lay_top = ftop - zvec[lay_fbot]
        if verbose:
            print(f"length per layer:\n- lay {lay_ftop}: {length_lay_top:.2f}"
                  f"\n- lay {lay_fbot}: {length_lay_bot:.2f}")
            
        if length_lay_top <= length_lay_bot:
            if verbose:
                print("selected layer:", lay_fbot)
            return lay_fbot
        
        elif length_lay_top > length_lay_bot:
            if verbose:
                print("selected layer:", lay_fbot)
            return lay_ftop
    
    raise ValueError('something is wrong with the input please contact Artesia')

    return np.nan

def get_zvec(x, y, gwf=None):
    """get a list with the vertical layer boundaries at a point in the model.

    Parameters
    ----------
    x : int or float
        x coördinate.
    y : int or float
        y coordinate.
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        modflow model with top and bottoms

    Raises
    ------
    NotImplementedError
        not all grid types are supported yet.

    Returns
    -------
    zvec : list
        list of vertical layer boundaries. length is nlay + 1.
    """
    
    if gwf is not None:
        if gwf.modelgrid.grid_type == 'structured':
            r,c = gwf.modelgrid.intersect(x,y)
            zvec = [gwf.modelgrid.top[r,c]] + [gwf.modelgrid.botm[i,r,c] for i in range(gwf.modelgrid.nlay)]
        elif gwf.modelgrid.grid_type == 'vertex':
            idx = gwf.modelgrid.intersect(x,y)
            zvec = [gwf.modelgrid.top[idx]] +[gwf.modelgrid.botm[i, idx] for i in range(gwf.modelgrid.nlay)]
        else:
            raise NotImplementedError(f'gridtype {gwf.modelgrid.grid_type} not (yet) implemented' )
        
    return zvec


@accessor.register_obscollection_accessor("gwobs")
class GwObsAccessor:

    def __init__(self, oc_obj):
        self._obj = oc_obj

    def set_filter_num(self, radius=1, xcol='x', ycol='y', if_exists='error',
                       add_to_meta=False):
        """This method computes the filternumbers based on the location of the
        observations. Then it sets the value of the filternumber:

            - in the ObsCollection dataframe
            - as the attribute of an Obs object
            - in the meta dictionary of the Obs object (only if add_to_meta is
            True)

        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        location with multiple filters. The filternr is based on the
        'onderkant_filter' attribute of the observations in such a way that
        the deepest filter has the highest filter number.

        Parameters
        ----------
        radius : int, optional
            max distance between two observations to be seen as one location, by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a filternr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the filternr is added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column filternr exists and if_exists='error' an error is raised
        """

        if self._obj['filternr'].dtype != np.number:
            self._obj['filternr'] = pd.to_numeric(
                self._obj['filternr'], errors='coerce')

        # check if column exists in obscollection
        if 'filternr' in self._obj.columns:
            if if_exists == 'error':
                raise RuntimeError(
                    "the column 'filternr' already exist, set if_exists='replace' to replace the current values")
            elif if_exists == 'replace':
                self._obj['filternr'] = np.nan
        else:
            self._obj['filternr'] = np.nan

        # ken filternummers toe aan peilbuizen die dicht bij elkaar staan
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, 'filternr']):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_filters = np.sqrt(
                    (self._obj[xcol] - x)**2 + (self._obj[ycol] - y)**2)
                dup_x = self._obj.loc[distance_to_other_filters < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(name, 'filternr', 1,
                                                  add_to_meta=add_to_meta)
                else:
                    dup_x2 = dup_x.sort_values(
                        'onderkant_filter', ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        self._obj._set_metadata_value(pb_dub, 'filternr', i + 1,
                                                      add_to_meta=add_to_meta)

    def set_filter_num_location(self, loc_col, radius=1, xcol='x', ycol='y',
                                if_exists='error', add_to_meta=False):
        """This method sets the filternr and locatie name of an observation
        point based on the locatie of the observations. When two or more
        filters are located close, defined by a radius, to eachother they get
        the same `locatie` and a different `filternr`.

        The value of the filternumber and the location are set:
            - in the ObsCollection dataframe
            - as the attribute of an Obs object
            - in the meta dictionary of the Obs object (only if add_to_meta is
            True)

        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        location with multiple filters. The filternr is based on the
        'onderkant_filter' attribute of the observations in such a way that
        the deepest filter has the highest filter number. The locatie is
        based on the named of the loc_col of the filter with the lowest
        filternr.

        Parameters
        ----------
        loc_col : str
            the column name with the names to use for the locatie
        radius : int, optional
            max distance between two observations to be seen as one location, by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a filternr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the filternr and location are added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column filternr exists and if_exists='error' an error is raised
        """

        # check if columns exists in obscollection
        if 'filternr' in self._obj.columns or 'locatie' in self._obj.columns:
            if if_exists == 'error':
                raise RuntimeError(
                    "the column 'filternr or locatie' already exist, set if_exists='replace' to replace the current values")
            elif if_exists == 'replace':
                self._obj['filternr'] = np.nan
                self._obj['locatie'] = np.nan
        else:
            self._obj['filternr'] = np.nan
            self._obj['locatie'] = np.nan

        # ken filternummers toe aan peilbuizen die dicht bij elkaar staan
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, 'filternr']):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_filters = np.sqrt(
                    (self._obj[xcol] - x)**2 + (self._obj[ycol] - y)**2)
                dup_x = self._obj.loc[distance_to_other_filters < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(name, 'filternr', 1,
                                                  add_to_meta=add_to_meta)
                    locatie = self._obj.loc[name, loc_col]
                    self._obj._set_metadata_value(name, 'locatie', locatie,
                                                  add_to_meta=add_to_meta)
                else:
                    dup_x2 = dup_x.sort_values(
                        'onderkant_filter', ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        if i == 0:
                            locatie = self._obj.loc[pb_dub, loc_col]
                        self._obj._set_metadata_value(pb_dub, 'filternr', i + 1,
                                                      add_to_meta=add_to_meta)
                        self._obj._set_metadata_value(pb_dub, 'locatie', locatie,
                                                      add_to_meta=add_to_meta)

    def get_modellayers(self, gwf, verbose=False):
        """Get the modellayer per observation. The layers can be obtained from
        the modflow model or can be defined in zgr.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        pd.Series with the modellayers of each observation
        """
        modellayers = []
        for o in self._obj.obs.values:
            modellayers.append(o.gwobs.get_modellayer_modflow(gwf, 
                                                              verbose=verbose))
            if verbose:
                print(o.name)

        modellayers = pd.Series(index=self._obj.index, data=modellayers,
                                name='modellayer')

        return modellayers
    
    


@accessor.register_obs_accessor("gwobs")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs
        
    def get_modellayer_modflow(self, gwf, left=-999, right=999, verbose=False):
        """Add modellayer to meta dictionary.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        int
            modellayer
        """
    
        zvec = get_zvec(self._obj.x, self._obj.y, gwf)
        
        modellayer = get_modellayer_from_filter(self._obj.bovenkant_filter, 
                                                self._obj.onderkant_filter, 
                                                zvec, 
                                                left=left, right=right,
                                                verbose=verbose)
        return modellayer
        



import os
import numpy as np
import pandas as pd
import xml.etree.ElementTree as etree
from lxml.etree import iterparse


def read_xml(fname, ObsClass, translate_dic={'locationId': 'locatie'},
             to_mnap=False, remove_nan=False, verbose=False):
    """read a FEWS XML-file with measurements, return list of ObsClass objects.

    Parameters
    ----------
    fname : str
        full path to file
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    remove_nan : boolean, optional
        remove nan values from measurements, flag information about the
        nan values is also lost
    verbose : boolean, optional
        print additional information to the screen (default is False).

    Returns
    -------
    list of ObsClass objects
        list of timeseries stored in ObsClass objects
    """
    tree = etree.parse(fname)
    root = tree.getroot()
    obs_list = []
    for i in range(len(root)):
        if root[i].tag.endswith('series'):
            series = {}
            date = []
            time = []
            events = []
            for j in range(len(root[i])):
                if root[i][j].tag.endswith('header'):
                    for k in range(len(root[i][j])):
                        prop = root[i][j][k].tag.split('}')[-1]
                        val = root[i][j][k].text
                        if prop == 'x' or prop == 'y' or prop == 'lat' or prop == 'lon':
                            val = float(val)
                        series[prop] = val
                        if verbose:
                            if prop == 'locationId':
                                print('read {}'.format(val))
                elif root[i][j].tag.endswith('event'):
                    date.append(root[i][j].attrib.pop('date'))
                    time.append(root[i][j].attrib.pop('time'))
                    events.append({**root[i][j].attrib})
            # combine events in a dataframe
            index = pd.to_datetime(
                [d + ' ' + t for d, t in zip(date, time)],
                errors="coerce")
            ts = pd.DataFrame(events, index=index, dtype=float)
            if remove_nan:
                ts.dropna(subset=['value'], inplace=True)
            if to_mnap:
                ts['stand_m_tov_nap'] = ts['value']

            if "x" in series.keys():
                x = series["x"]
            else:
                x = np.nan
            if "y" in series.keys():
                y = series["y"]
            else:
                y = np.nan

            for key, item in translate_dic.items():
                series[item] = series.pop(key)

            obs_list.append(ObsClass(ts, name=series['locatie'],
                                     locatie=series['locatie'],
                                     x=x, y=y, meta=series))
    return obs_list


def iterparse_pi_xml(fname, ObsClass, translate_dic={'locationId': 'locatie'},
                     locationIds=None, return_events=True,
                     keep_flags=(0, 1), return_df=False,
                     tags=('series', 'header', 'event'),
                     skip_errors=True, verbose=False):
    """read a FEWS XML-file with measurements, memory efficient.

    Parameters
    ----------
    fname : str
        full path to file
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    locations : tuple or list of str, optional
        list of locationId's to read from XML file, others are skipped.
        If None (default) all locations are read.
    return_events : bool, optional
        return all event-information in a DataFrame per location, instead of
        just a Series (defaults to False). Overrules keep_flags kwarg.
    keep_flags : list of ints, optional
        keep the values with these flags (defaults to 0 and 1). Only used
        when return_events is False.
    tags : list of strings, optional
        Select the tags to be parsed. Defaults to series, header and event
    return_df : bool, optional
        return a DataFame with the data, instead of two lists (defaults to
        False)
    skip_errors: bool, optional
        if True, continue after error, else raise error

    Returns
    -------

    if return_df == True:
        df : pandas DataFrame
            a DataFrame containing the metadata and the series
    else:
        header_list : list of dictionaries
            list of metadata
        series_list : list of pandas Series
            list of timeseries
    """

    tags = ['{{http://www.wldelft.nl/fews/PI}}{}'.format(tag) for tag in tags]

    context = iterparse(fname, tag=tags)

    header_list = []
    series_list = []

    keep_flags = [str(flag) for flag in keep_flags]

    try:
        for _, element in context:
            if element.tag.endswith("header"):
                header = {}
                for h_attr in element:
                    tag = h_attr.tag.replace("{{{0}}}".format(
                        "http://www.wldelft.nl/fews/PI"), "")
                    # if specific locations are provided only read those
                    if locationIds is not None and tag.startswith("locationId"):
                        loc = h_attr.text
                        if loc not in locationIds:
                            continue
                    if h_attr.text is not None:
                        header[tag] = h_attr.text
                    elif len(h_attr.attrib) != 0:
                        header[tag] = {**h_attr.attrib}
                    else:
                        header[tag] = None
                    if verbose and tag.startswith("locationId"):
                        print("reading {}".format(header[tag]))
                events = []
            elif element.tag.endswith("event"):
                # if specific locations are provided only read those
                if locationIds is not None:
                    if loc not in locationIds:
                        continue
                events.append({**element.attrib})
            elif element.tag.endswith('series'):
                # if specific locations are provided only read those
                if locationIds is not None:
                    if loc not in locationIds:
                        continue
                if len(events) == 0:
                    if return_events:
                        s = pd.DataFrame()
                    else:
                        s = pd.Series()
                else:
                    df = pd.DataFrame(events)
                    df.index = pd.to_datetime(
                        [d + " " + t for d, t in zip(df['date'], df['time'])],
                        errors="coerce")
                    df.drop(columns=["date", "time"], inplace=True)
                    if return_events:
                        df['value'] = pd.to_numeric(
                            df['value'], errors="coerce")
                        df['flag'] = pd.to_numeric(df['flag'])
                        s = df
                    else:
                        mask = df['flag'].isin(keep_flags)
                        s = pd.to_numeric(
                            df.loc[mask, 'value'], errors="coerce")

                for key, item in translate_dic.items():
                    header[item] = header.pop(key)

                o = ObsClass(s, name=header['locatie'],
                             locatie=header['locatie'],
                             meta=header)
                header_list.append(header)
                series_list.append(o)

            # Free memory.
            element.clear()
            # while element.getprevious() is not None:
            #    del element.getparent()[0]

    except Exception as e:
        if skip_errors:
            print("ERROR! Skipped {}".format(fname))
        else:
            raise(e)

    if return_df:
        for h, s in zip(header_list, series_list):
            h['series'] = s
        return pd.DataFrame(header_list)
    else:
        return header_list, series_list


def write_pi_xml(obs_coll, fname, timezone=1.0, version="1.24"):
    """Write PiTimeSeries object to PI-XML file.

    Parameters
    ----------
    fname: path
        path to XML file to be written
    """

    assert fname.endswith(
        ".xml"), "Output file should have '.xml' extension!"

    # first line of XML file
    line0 = '<?xml version="1.0" encoding="UTF-8"?>\n'

    # some definitions for timeseries XML file
    NS = r"http://www.wldelft.nl/fews/PI"
    FS = r"http://www.wldelft.nl/fews/fs"
    XSI = r"http://www.w3.org/2001/XMLSchema-instance"
    schemaLocation = (
        r"http://fews.wldelft.nl/schemas/version1.0"
        r"/Pi-schemas/pi_timeseries.xsd"
    )
    timeseriesline = ('<TimeSeries xmlns="{NS}" xmlns:xsi="{XSI}" '
                      'xsi:schemaLocation="{NS} {schema}" version="{version}" '
                      'xmlns:fs="{FS}">\n')

    # line templates
    paramline = "<{tag}>{param}</{tag}>\n"

    # write file
    with open(fname, "w") as f:
        f.write(line0)
        f.write(
            timeseriesline.format(
                NS=NS, FS=FS, XSI=XSI, schema=schemaLocation, version=version
            )
        )
        tzline = "\t" + \
            paramline.format(tag="timeZone", param=timezone)
        f.write(tzline)

        for o in obs_coll.obs:
            # start series
            start = "\t" + "<series>\n"
            f.write(start)
            # write header
            hlines = []
            hstart = 2 * "\t" + "<header>\n"
            hlines.append(hstart)
            for htag, hval in o.meta.items():
                if htag.endswith("Date"):
                    try:
                        hdate = hval.strftime("%Y-%m-%d")
                        htime = hval.strftime("%H:%M:%S")
                    except AttributeError as e:
                        if htag.startswith("start"):
                            hdate = o.index[0].strftime("%Y-%m-%d")
                            htime = o.index[0].strftime("%H:%M:%S")
                        elif htag.startswith("end"):
                            hdate = o.index[-1].strftime("%Y-%m-%d")
                            htime = o.index[-1].strftime("%H:%M:%S")
                        else:
                            raise(e)
                    hline = '<{tag} date="{date}" time="{time}"/>\n'.format(
                        tag=htag, date=hdate, time=htime
                    )
                elif htag.endswith("timeStep"):
                    hline = '<{tag} unit="{unit}"/>\n'.format(
                        tag=htag, unit=hval)
                else:
                    hline = paramline.format(tag=htag, param=hval)
                hlines.append(3 * "\t" + hline)
            hlines.append(2 * "\t" + "</header>\n")
            f.writelines(hlines)

            # write timeseries
            dates = o.reset_index()["index"].apply(
                lambda s: pd.datetime.strftime(s, "%Y-%m-%d")
            )
            times = o.reset_index()["index"].apply(
                lambda s: pd.datetime.strftime(s, "%H:%M:%S")
            )
            # set date and time attributes
            events = (
                2 * "\t"
                + '<event date="'
                + dates.values
                + '" time="'
                + times.values
            )
            # loop through columns and add to event
            for icol in o.columns:
                val = o[icol].astype(str)
                events += (
                    '" {}="'.format(icol)
                    + val.values
                )
            # close event
            events += '"/>\n'
            # write to file
            f.writelines(events)
            # end series
            f.write("\t" + "</series>\n")
        # end Timeseries
        f.write("</TimeSeries>\n")


def parse_xml_filelist(fnames, ObsClass, directory=None, locations=None,
                       translate_dic={'locationId': 'locatie'},
                       to_mnap=False, remove_nan=False, verbose=False,
                       low_memory=True):
    obs_list = []
    nfiles = len(fnames)
    for j, ixml in enumerate(fnames):

        # print message
        if verbose:
            print("{0}/{1} read {2}".format(j + 1, nfiles, ixml))

        # join directory to filename if provided
        if directory is None:
            fullpath = ixml
        else:
            fullpath = os.path.join(directory, ixml)

        # selection of xml parse method
        if low_memory:
            olist = read_xml(fullpath,
                             ObsClass=ObsClass,
                             translate_dic=translate_dic,
                             to_mnap=to_mnap,
                             remove_nan=remove_nan,
                             verbose=False)
        else:
            _, olist = iterparse_pi_xml(fullpath, ObsClass,
                                        translate_dic=translate_dic,
                                        locationIds=locations,
                                        verbose=verbose)
        obs_list += olist
    return obs_list



"""Module with a number of observation classes.

The Obs class is a subclass of a pandas DataFrame with
additional attributes and methods. The specific classes (GroundwaterObs,
WaterlvlObs, ...) are subclasses of the Obs class.

The subclasses of a dataframe can have additional attributes and methods.
Additional attributes have to be defined in the '_metadata' attribute. In order
to keep the subclass methods and attributes when selecting or slicing an object
you need the '_constructor' method.


More information about subclassing pandas DataFrames can be found here:
http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
"""

import warnings
import numpy as np
from pandas import DataFrame


class Obs(DataFrame):
    """class for point observations.

    An Obs object is a subclass of a pandas.DataFrame and allows for additional
    attributes and methods.
    pandas can be found here:
    http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas

    Parameters
    ----------
    x : int or float
        x coordinate of observation point
    y : int or float
        y coordinate of observation point
    name : str
        name
    meta : dictionary
        metadata
    filename : str
        filename with data of observation point
    """
    # temporary properties
    _internal_names = DataFrame._internal_names + ['none']
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ['x', 'y', 'name', 'meta', 'filename']

    def __init__(self, *args, **kwargs):
        """constructor of Obs class.

        *args must be input for the pandas.DataFrame constructor,
        **kwargs can be one of the attributes listed in _metadata or
        keyword arguments for the constructor of a pandas.DataFrame.
        """
        self.x = kwargs.pop('x', np.nan)
        self.y = kwargs.pop('y', np.nan)
        self.name = kwargs.pop('name', '')
        self.meta = kwargs.pop('meta', {})
        self.filename = kwargs.pop('filename', '')

        super(Obs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return Obs

    def to_collection_dict(self, include_meta=False):
        """get dictionary with registered attributes and their values of an Obs
        object.

        This method can be used to create a dataframe from a collection
        of Obs objects.

        Parameters
        ----------
        include_meta : boolean, optional
            include the meta dictionary in the collection dictionary,
            default is false

        Returns
        -------
        d : dictionary
            dictionary with Obs information
        """
        
        attrs = self._metadata.copy()
        if not include_meta:
            attrs.remove('meta')
        
        d = {}
        for att in attrs:
            d[att] = getattr(self, att)

        d['obs'] = self

        return d


class GroundwaterObs(Obs):
    """class for groundwater quantity point observations.

    Subclass of the Obs class. Can have the following attributes:
        - locatie: 2 filters at one piezometer should have the same 'locatie'
        - filternr: 2 filters at one piezometer should have a different 'filternr'.
        a higher filter number is preferably deeper than a lower filter number.
        - bovenkant_filter: top op the filter in m NAP
        - onderkant_filter: bottom of the filter in m NAP
        - maaiveld: surface level in m NAP
        - meetpunt: ? in m NAP
        - metadata_available: boolean indicating if metadata is available for
        the measurement point.
    """

    _metadata = Obs._metadata + \
        ['locatie', 'filternr',
         'bovenkant_filter', 'onderkant_filter',
         'maaiveld', 'meetpunt', 'metadata_available'
         ]

    def __init__(self, *args, **kwargs):
        """
        *args must be input for the pandas.DataFrame constructor,
        **kwargs can be one of the attributes listed in _metadata or
        keyword arguments for the constructor of a pandas.DataFrame.

        if the pandas.DataFrame has a column 'stand_m_tov_nap' a lot of
        plotting and other methods will work automatically without changing
        the default arguments.
        """
        self.locatie = kwargs.pop('locatie', '')
        self.filternr = kwargs.pop('filternr', '')
        self.maaiveld = kwargs.pop('maaiveld', np.nan)
        self.meetpunt = kwargs.pop('meetpunt', np.nan)
        self.bovenkant_filter = kwargs.pop('bovenkant_filter', np.nan)
        self.onderkant_filter = kwargs.pop('onderkant_filter', np.nan)
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(GroundwaterObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GroundwaterObs

    @classmethod
    def from_dino(cls, fname=None, location=None, filternr=1.,
                  tmin="1900-01-01", tmax="2040-01-01",
                  **kwargs):
        """download dino data from the server.

        Parameters
        ----------
        fname : str, optional
            dino csv filename
        location : str, optional
            location of the peilbuis, i.e. B57F0077
        filternr : float, optional
            filter_nr of the peilbuis, i.e. 1.
        tmin : str
            start date in format YYYY-MM-DD
        tmax : str
            end date in format YYYY-MM-DD
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv if
            fname is not None and otherwise to io_dino.findMeetreeks
        """
        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_dino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)

        elif location is not None:
            measurements, meta = io_dino.download_dino_groundwater(location,
                                                                   filternr,
                                                                   tmin, tmax,
                                                                   **kwargs)

            if meta['metadata_available']:
                return cls(measurements, meta=meta,
                           x=meta.pop('x'), y=meta.pop('y'),
                           onderkant_filter=meta.pop('onderkant_filter'),
                           bovenkant_filter=meta.pop('bovenkant_filter'),
                           name=meta.pop('name'),
                           metadata_available=meta.pop('metadata_available'),
                           locatie=meta.pop('locatie'),
                           maaiveld=meta.pop('maaiveld'),
                           filternr=meta.pop('filternr'))
            else:
                return cls(measurements, meta=meta)
        else:
            raise ValueError(
                'specify fname or location to obtain groundwater heads')

    @classmethod
    def from_dino_server(cls, location, filternr=1.,
                         tmin="1900-01-01", tmax="2040-01-01",
                         **kwargs):
        """download dino data from the server.

        Parameters
        ----------
        location : str
            location of the peilbuis, i.e. B57F0077
        filternr : float
            filter_nr of the peilbuis, i.e. 1.0
        tmin : str
            start date in format YYYY-MM-DD
        tmax : str
            end date in format YYYY-MM-DD
        kwargs : key-word arguments
            these arguments are passed to dino.findMeetreeks functie
        """
        warnings.warn(
            "this method will be removed in future versions, use from_dino instead", DeprecationWarning)

        from .io import io_dino
        measurements, meta = io_dino.download_dino_groundwater(location,
                                                               filternr,
                                                               tmin, tmax,
                                                               **kwargs)

        if meta['metadata_available']:
            return cls(measurements, meta=meta,
                       x=meta.pop('x'), y=meta.pop('y'),
                       onderkant_filter=meta.pop('onderkant_filter'),
                       bovenkant_filter=meta.pop('bovenkant_filter'),
                       name=meta.pop('name'),
                       locatie=meta.pop('locatie'),
                       maaiveld=meta.pop('maaiveld'),
                       filternr=meta.pop('filternr'))
        else:
            return cls(measurements, meta=meta)

    @classmethod
    def from_dino_file(cls, fname=None, **kwargs):
        """read a dino csv file.

        Parameters
        ----------
        name : str, optional
            name of the peilbuis, i.e. B57F0077
        fname : str, optional
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv
        """
        warnings.warn(
            "this method will be removed in future versions, use from_dino instead", DeprecationWarning)

        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_dino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)
        else:
            raise ValueError(
                'specify either the name or the filename of the measurement point')

    @classmethod
    def from_artdino_file(cls, fname=None, **kwargs):
        """read a dino csv file (artdiver style).

        Parameters
        ----------
        name : str, optional
            name of the peilbuis, i.e. B57F0077
        fname : str, optional
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv
        """
        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_artdino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)
        else:
            raise ValueError('specify either the name or the filename of the '
                             'measurement point!')

    @classmethod
    def from_wiski(cls, fname, **kwargs):
        """[summary]

        Parameters
        ----------
        fname : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """
        from .io import io_wiski
        data, metadata = io_wiski.read_wiski_file(fname, **kwargs)

        return cls(data, meta=metadata, **metadata)

    @classmethod
    def from_pystore_item(cls, item):
        """Create GroundwaterObs DataFrame from Pystore item.

        Parameters
        ----------
        item : pystore.item.Item
            Pystore item

        Returns
        -------
        GroundwaterObs
            GroundwaterObs DataFrame
        """

        df = item.to_pandas()
        try:
            x = item.metadata["x"]
            y = item.metadata["y"]
        except KeyError:
            x = np.nan
            y = np.nan
        item.metadata["datastore"] = item.datastore
        return cls(df, x=x, y=y, meta=item.metadata)


class GroundwaterQualityObs(Obs):
    """class for groundwater quality (grondwatersamenstelling) point
    observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + \
        ['locatie', 'filternr', 'maaiveld', 'metadata_available']

    def __init__(self, *args, **kwargs):

        self.locatie = kwargs.pop('locatie', '')
        self.filternr = kwargs.pop('filternr', '')
        self.maaiveld = kwargs.pop('maaiveld', np.nan)
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(GroundwaterQualityObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GroundwaterQualityObs

    @classmethod
    def from_dino(cls, fname, **kwargs):
        """read ad dino file with groundwater quality data.

        Parameters
        ----------
        fname : str
            dino txt filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_quality_txt
        """
        from .io import io_dino
        measurements, meta = io_dino.read_dino_groundwater_quality_txt(
            fname, **kwargs)

        return cls(measurements, meta=meta, **meta)


class WaterlvlObs(Obs):
    """class for water level point observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['locatie', 'metadata_available']

    def __init__(self, *args, **kwargs):

        self.locatie = kwargs.pop('locatie', '')
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(WaterlvlObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return WaterlvlObs

    @classmethod
    def from_dino(cls, fname, **kwargs):
        """read a dino file with waterlvl data.

        Parameters
        ----------
        fname : str
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_waterlvl_csv
        """
        from .io import io_dino
        measurements, meta = io_dino.read_dino_waterlvl_csv(fname, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_waterinfo(cls, fname, **kwargs):
        """Read data from waterinfo csv-file or zip.

        Parameters
        ----------
        fname : str
            path to file (file can zip or csv)

        Returns
        -------
        df : WaterlvlObs
            WaterlvlObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import io_waterinfo
        df, metadata = io_waterinfo.read_waterinfo_file(fname,
                                                        return_metadata=True)
        return cls(df, meta=metadata, **metadata)


class ModelObs(Obs):
    """class for model point results.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['model']

    def __init__(self, *args, **kwargs):

        self.model = kwargs.pop('model', '')

        super(ModelObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return ModelObs


class KnmiObs(Obs):
    """class for KNMI timeseries.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['station']

    def __init__(self, *args, **kwargs):

        self.station = kwargs.pop('station', np.nan)

        super(KnmiObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return KnmiObs

    @classmethod
    def from_knmi(cls, stn, variable, startdate=None, enddate=None,
                  fill_missing_obs=True, verbose=False):
        from .io import io_knmi

        ts, meta = io_knmi.get_knmi_timeseries_stn(stn, variable,
                                                   startdate, enddate,
                                                   fill_missing_obs,
                                                   verbose=verbose)
        
        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])

    @classmethod
    def from_nearest_xy(cls, x, y, variable, startdate=None, enddate=None,
                        fill_missing_obs=True, verbose=False):
        from .io import io_knmi

        ts, meta = io_knmi.get_knmi_timeseries_xy(x, y, variable,
                                                  startdate, enddate,
                                                  fill_missing_obs,
                                                  verbose=verbose)

        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])

    @classmethod
    def from_obs(cls, obs, variable, startdate=None, enddate=None,
                 fill_missing_obs=True, verbose=False):

        from .io import io_knmi

        x = obs.x
        y = obs.y

        if startdate is None:
            startdate = obs.index[0]
        if enddate is None:
            enddate = obs.index[-1]

        ts, meta = io_knmi.get_knmi_timeseries_xy(x, y, variable,
                                                  startdate, enddate,
                                                  fill_missing_obs,
                                                  verbose=verbose)

        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])



from setuptools import setup, find_packages

try:
    import pypandoc
    l_d = pypandoc.convert('README.md')
except:
    l_d = ''

# Get the version.
version = {}
with open("version.py") as fp:
    exec(fp.read(), version)
setup(
    name='hydropandas',
    version=version['__version__'],
    description='hydropandas module by Artesia',
    long_description=l_d,
	long_description_content_type="text/markdown",
    url='https://github.com/ArtesiaWater/hydropandas',
    author='Artesia',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    platforms='Windows, Mac OS-X',
    install_requires=['numpy>=1.15', 'matplotlib>=3.0',
                      'pandas>=0.24', 'scipy>=1.2', 
					  'geopandas >= 0.6.0', 'zeep >= 3.4.0',
					  'bokeh >= 2.0.2','branca >= 0.3.1',
					  'folium >= 0.8.2', 'pastas >= 0.13.0'],
    packages=find_packages(exclude=[]),
    package_data={"hydropandas": ["data/*.json"]},
    include_package_data=True
)



from .obs_collection import ObsCollection
from .observation import (GroundwaterObs, GroundwaterQualityObs, KnmiObs,
                          ModelObs, WaterlvlObs)
from .extensions import geo as _geo
from .extensions import gwobs as _gwobs
from .extensions import plots as _plots
from .extensions import stats as _stats
